from __future__ import annotations

import operator as op
from pathlib import Path
from typing import (
    Any,
    NamedTuple,
    Optional,
    Sequence,
    TypeAlias,
    Union,
    cast,
    Callable,
    TypeGuard,
)
from .config_var_def import ConfigVarDef
from numbers import Number
from .utils import _undef
import json
import inspect

PathType = Union[Path, ConfigVarDef, "ConfigVarExpression"]
""" Type that support Path operation """
PathOperand = Union[str, PathType]
""" operand that can be combined with a PathType in a Path operation """
NumericType = Union[Number, ConfigVarDef, "ConfigVarExpression"]
""" Type that support numeric operations """
NumericOperand = NumericType
""" operand that can be combined with a NumericType in a numeric operation """
StringType = Union[str, ConfigVarDef, "ConfigVarExpression"]
""" Type that support string operations """
StringOperand = StringType
""" operand that can be combined with a StringType in a string operation """
OperandType: TypeAlias = Union[
    PathOperand,
    NumericOperand,
    StringOperand,
]
""" Types that are supported as operand """


UnaryOperator = Callable[[OperandType], Any]
# Unary Operator type
BinaryOperator = Callable[[OperandType, OperandType], Any]
# Binary Operator type


def _is_operand(value: Any) -> TypeGuard[OperandType]:
    """Return True when value is a runtime-valid expression operand.

    Runtime checks cannot use ``isinstance(value, OperandType)`` safely because
    OperandType includes forward references through typing aliases.
    """
    return isinstance(value, (str, Path, Number, ConfigVarDef, ConfigVarExpression))


class ExpressionDecorator(NamedTuple):
    """Decorator of an expression node with the value resulting of applying
    the operator on  operands and the type of the result.
    """

    value: Any
    type_info: type


def config_path_expression(
    left: PathOperand, *others: PathOperand
) -> ConfigVarExpression:
    """Factory function to create a ConfigVarExpression from PathOperands.

    Used :
        - By Jinja filter and function 'p' and 'P'.
        - To provide backward compatibility with deprecated SegmentedPath.
    """

    if isinstance(left, ConfigVarExpression):
        left_expr = left
    elif isinstance(left, Path):
        left_expr = left
    elif isinstance(left, str):
        left_expr = Path(left)
    elif isinstance(left, ConfigVarDef):
        left_expr = left
    else:
        raise TypeError(
            f"ConfigVarDef {left.name} with type {left.type_info} is not compatible with Path operation"
        )

    if len(others) > 0:
        right = (
            config_path_expression(others[0], *others[1:])
            if len(others) > 1
            else others[0]
        )
        return ConfigVarExpression(left_expr, right, op.truediv, is_unary=False)
    else:
        return ConfigVarExpression(
            left_expr, None, lambda x: x, is_unary=True  # identity function
        )


class ConfigVarExpression:
    """Class representing an expression combining several configuration variables with different operators."""

    _expression: ConfigVarOperator
    """ The expression tree representing the operations to perfom at resolution time.
    """
    _resolved_value: Any = _undef
    """ The resolved value of this expression, set at resolution time. """
    _resolved_type: Optional[type] = None
    """ The resolved type of this expression, set at resolution time. """

    def __init__(
        self,
        left: OperandType,
        right: Optional[OperandType],
        operator: UnaryOperator | BinaryOperator,
        *,
        is_unary: bool = False,
    ) -> None:
        """Create a ConfigVarExpression that can be used in ConfigVar operation."""
        self._expression = ConfigVarOperator(
            left=left,
            right=right,
            operator=operator,
            is_unary=is_unary,
        )

    def resolve(self, substitutor: Callable[[Any], Any]) -> Any:
        """Resolve this expression to a concrete value by resolving the expression tree with the provided substitutor."""
        if self._resolved_value is not _undef:
            return self._resolved_value
        value, type_info = self._expression.resolve(substitutor)
        self._resolved_value = value
        self._resolved_type = type_info
        return value

    def _binary(
        self,
        other: OperandType,
        operator_fn: BinaryOperator,
    ) -> ConfigVarExpression:
        """Helper method used to create binary operation with self on the left and other on the right of the operator"""
        if not _is_operand(other):
            return NotImplemented
        return ConfigVarExpression(self, other, operator_fn)

    def _rbinary(
        self,
        other: OperandType,
        operator_fn: BinaryOperator,
    ) -> ConfigVarExpression:
        """Helper method used to create binary operation with self on the right and other on the left of the operator"""
        if not _is_operand(other):
            return NotImplemented
        return ConfigVarExpression(other, self, operator_fn)

    def __add__(self, other: OperandType) -> ConfigVarExpression:
        """Provide support for + operator between this ConfigVarExpression and another supported operand, returning a new ConfigVarExpression
        representing the addition operation."""
        if not _is_operand(other):
            return NotImplemented
        return self._binary(other, op.add)

    def __radd__(self, other: OperandType) -> ConfigVarExpression:
        """Provide support for + operator between another supported operand and this ConfigVarExpression, returning a new ConfigVarExpression
        representing the addition operation."""
        if not _is_operand(other):
            return NotImplemented
        return self._rbinary(other, op.add)

    def __sub__(self, other: OperandType) -> ConfigVarExpression:
        """Provide support for - operator between this ConfigVarExpression and another supported operand, returning a new ConfigVarExpression
        representing the subtraction operation."""
        if not _is_operand(other):
            return NotImplemented
        return self._binary(other, op.sub)

    def __rsub__(self, other: OperandType) -> ConfigVarExpression:
        """Provide support for - operator between another supported operand and this ConfigVarExpression, returning a new ConfigVarExpression
        representing the subtraction operation."""
        if not _is_operand(other):
            return NotImplemented
        return self._rbinary(other, op.sub)

    def __mul__(self, other: OperandType) -> ConfigVarExpression:
        """Provide support for * operator between this ConfigVarExpression and another supported operand returning a new ConfigVarExpression
        representing the multiplication operation."""
        if not _is_operand(other):
            return NotImplemented
        return self._binary(other, op.mul)

    def __rmul__(self, other: OperandType) -> ConfigVarExpression:
        """Provide support for * operator between another supported operand and this ConfigVarExpression, returning a new ConfigVarExpression
        representing the multiplication operation."""
        if not _is_operand(other):
            return NotImplemented
        return self._rbinary(other, op.mul)

    def __truediv__(self, other: OperandType) -> ConfigVarExpression:
        """Provide support for / operator between another supported operand and this ConfigVarExpression, returning a new ConfigVarExpression
        representing the division operation."""
        if not _is_operand(other):
            return NotImplemented
        return self._binary(other, op.truediv)

    def __rtruediv__(self, other: OperandType) -> ConfigVarExpression:
        """Provide support for / operator between another supported operand and this ConfigVarExpression, returning a new ConfigVarExpression
        representing the division operation."""
        if not _is_operand(other):
            return NotImplemented
        return self._rbinary(other, op.truediv)

    def __floordiv__(self, other: OperandType) -> ConfigVarExpression:
        """Provide support for // operator between this ConfigVarExpression and another supported operand, returning a new ConfigVarExpression
        representing the floor division operation."""
        if not _is_operand(other):
            return NotImplemented
        return self._binary(other, op.floordiv)

    def __rfloordiv__(self, other: OperandType) -> ConfigVarExpression:
        """Provide support for // operator between another supported operand and this ConfigVarExpression, returning a new ConfigVarExpression
        representing the floor division operation."""
        if not _is_operand(other):
            return NotImplemented
        return self._rbinary(other, op.floordiv)

    def __mod__(self, other: OperandType) -> ConfigVarExpression:
        """Provide support for % operator between this ConfigVarExpression and another supported operand, returning a new ConfigVarExpression
        representing the modulo operation."""
        if not _is_operand(other):
            return NotImplemented
        return self._binary(other, op.mod)

    def __rmod__(self, other: OperandType) -> ConfigVarExpression:
        """Provide support for % operator between another supported operand and this ConfigVarExpression, returning a new ConfigVarExpression
        representing the modulo operation."""
        if not _is_operand(other):
            return NotImplemented
        return self._rbinary(other, op.mod)

    def __pow__(self, other: OperandType) -> ConfigVarExpression:
        """Provide support for ** operator between this ConfigVarExpression and another supported operand, returning a new ConfigVarExpression
        representing the exponentiation operation."""
        if not _is_operand(other):
            return NotImplemented
        return self._binary(other, op.pow)

    def __rpow__(self, other: OperandType) -> ConfigVarExpression:
        """Provide support for ** operator between another supported operand and this ConfigVarExpression, returning a new ConfigVarExpression
        representing the exponentiation operation."""
        if not _is_operand(other):
            return NotImplemented
        return self._rbinary(other, op.pow)

    def __neg__(self) -> ConfigVarExpression:
        """Provide support for the unary - operator on this ConfigVarExpression, returning a new ConfigVarExpression
        representing the negation operation."""
        return ConfigVarExpression(
            self,
            None,
            cast(UnaryOperator, op.neg),
            is_unary=True,
        )

    def __pos__(self) -> ConfigVarExpression:
        """Provide support for the unary + operator on this ConfigVarExpression, returning a new ConfigVarExpression
        representing the positive operation."""
        return ConfigVarExpression(
            self,
            None,
            cast(UnaryOperator, op.pos),
            is_unary=True,
        )

    def __abs__(self) -> ConfigVarExpression:
        """Provide support for the abs() function on this ConfigVarExpression, returning a new ConfigVarExpression
        representing the absolute value operation."""
        return ConfigVarExpression(
            self,
            None,
            cast(UnaryOperator, op.abs),
            is_unary=True,
        )

    def __json__(self) -> dict[str, Any]:
        """Json representation of the tree representing the expression."""
        return self._expression.__json__()

    # def __str__(self) -> str:
    #     """String representation of this ConfigVarExpression.
    #     If the expression is resolved, return the resolved value and type, otherwise return the expression tree
    #     """
    #     if self._resolved_value is not _undef:
    #         return f"{self._resolved_value}/{self._resolved_type}"
    #     return json.dumps(self.__json__(), indent=2)


class ConfigVarOperator:
    """Class representing a numeric operation that can be applied to ConfigVarExpression values and that can
       be resolved with variable resolution before applying the operator.
    Note: Resolution implies jinja template rendering and resolution of the value.
    """

    _right: Any
    """ right operand"""
    _left: Any
    """ left operand"""
    _unary_operator: Optional[UnaryOperator]
    """ unary operator"""
    _binary_operator: Optional[BinaryOperator]
    """ binary operator"""
    _is_unary: bool
    """ flag indicating if the operator is unary """
    _type_info: Optional[type]
    """ type of the result of this operation, computed at resolution time based on the operator and the type of the operands """
    _resolved_value: Any
    """ resolved value of this operation, computed at resolution time by applying the operator on the resolved operands """

    def __init__(
        self,
        left: Any,
        right: Any | None,
        operator: UnaryOperator | BinaryOperator,
        *,
        is_unary: bool = False,
    ) -> None:
        """Create a SegmentedVarOperator representing either a unary or binary operation.

        IMPORTANT: we use Any for left and right in order to avoid confusing the type checker with the ConfigVarDef
        assignement that would be interpreted as a property assignement (ConfigVarDef defines __set__()) not compatible
        with the OperandType.

        """
        self._left = left
        self._right = right
        self._is_unary = is_unary
        self._unary_operator = cast(UnaryOperator, operator) if is_unary else None
        self._binary_operator = None if is_unary else cast(BinaryOperator, operator)

    @staticmethod
    def compute_type(
        operator: UnaryOperator | BinaryOperator,
        left_type: type,
        right_type: Optional[type] = None,
    ) -> type:
        """Compute the type of the result of applying the operator on left and right operator.

        right is None for unary operattion.

        The operators that pose problem are:
        - division: it can be used in Path operations and in numeric operations.
        - addition: it can be used in string concatenation and in numeric addition.
        For all other operators we return numbers.Number as the result type.

        For division if one operand is a Number , we return numbers.Number as the result type, otherwise Path.
        For addition if one operand is a Number , we return numbers.Number as the result type, otherwise str.
        """
        if right_type is None:
            if operator in (op.neg, op.pos, op.abs):
                return Number
            else:
                return left_type
        if operator == op.truediv:
            if (
                isinstance(left_type, type)
                and issubclass(left_type, Number)
                or isinstance(right_type, type)
                and issubclass(right_type, Number)
            ):
                return Number
            else:
                return Path
        if operator == op.add:
            if (
                isinstance(left_type, type)
                and issubclass(left_type, Number)
                or isinstance(right_type, type)
                and issubclass(right_type, Number)
            ):
                return Number
            else:
                return str
        return Number

    def _decorate_operand(
        self,
        operand: OperandType,
        substitutor: Callable[[Any], Any],
    ) -> ExpressionDecorator:
        if operand is None:
            return ExpressionDecorator(None, type(None))
        elif isinstance(operand, ConfigVarExpression):
            value, type_info = operand._expression.resolve(substitutor)
            return ExpressionDecorator(value, type_info)
        elif isinstance(operand, ConfigVarDef):
            value = substitutor(operand)
            type_info = operand.type_info
            return ExpressionDecorator(value, type_info)
        elif isinstance(operand, str):
            value = substitutor(operand)
            type_info = type(value)
            return ExpressionDecorator(value, type_info)
        else:
            value = substitutor(operand)
            type_info = type(value)
            return ExpressionDecorator(value, type_info)

    def resolve(self, substitutor: Callable[[Any], Any]) -> ExpressionDecorator:
        """Resolve and evaluate this operator tree to a concrete value.

        The resolution is done by decorating the expression tree with the intermediate values and types of the operands and then computing the result and its type
        based on the decorated operands and the operator.
        We use a substitutor function to resolve the value of ConfigVarDef and to render the jinja template if the operand is a string.
        """

        # first decorate the tree with type information and intermediate values.
        right_value, right_type = None, None
        left_value, left_type = self._decorate_operand(self._left, substitutor)
        if not self._is_unary:
            right_value, right_type = self._decorate_operand(self._right, substitutor)

        if self._is_unary:
            if self._unary_operator is None:
                raise ValueError("Unary operation requires a unary operator")
            # we don't check the type of operand because we want to let the interpreter do the casting if possible
            res_type = self.compute_type(self._unary_operator, left_type)
            res = self._unary_operator(left_value)
            return ExpressionDecorator(res, res_type)

        if (
            self._right is None
            or self._binary_operator is None
            or right_value is None
            or right_type is None
        ):
            raise ValueError(
                "Binary operation requires a right operand and a binary operator"
            )

        res_type = self.compute_type(self._binary_operator, left_type, right_type)
        if self._binary_operator == op.add:
            if isinstance(res_type, type) and issubclass(res_type, str):
                res = f"{str(left_value)}{str(right_value)}"
            else:
                # we don't check the type of operand because we want to let the interpreter do the casting if possible
                res = self._binary_operator(left_value, right_value)
        elif self._binary_operator == op.truediv:
            if isinstance(res_type, type) and issubclass(res_type, Path):
                res = Path(str(left_value), str(right_value))
            else:
                # we don't check the type of operand because we want to let the interpreter do the casting if possible
                res = self._binary_operator(left_value, right_value)
        else:
            # we don't check the type of operand because we want to let the interpreter do the casting if possible
            res = self._binary_operator(left_value, right_value)
        return ExpressionDecorator(res, res_type)

    @staticmethod
    def _operator_str(operator: UnaryOperator | BinaryOperator) -> str:
        """Helper method to get a string representation of an operator function."""
        if operator == op.add:
            return "+"
        elif operator == op.sub:
            return "-"
        elif operator == op.mul:
            return "*"
        elif operator == op.truediv:
            return "/"
        elif operator == op.floordiv:
            return "//"
        elif operator == op.mod:
            return "%"
        elif operator == op.pow:
            return "**"
        elif operator == op.neg:
            return "-"
        elif operator == op.pos:
            return "+"
        elif operator == op.abs:
            return "abs"
        else:
            if callable(operator) and getattr(operator, "__name__", "") == "<lambda>":
                try:
                    return inspect.getsource(operator).strip()
                except (OSError, TypeError):
                    pass
            return str(operator)

    def __json__(self) -> dict[str, Any]:
        """JSON representation of this ConfigVarOperator."""
        operator_str: str = ""
        if self._is_unary and self._unary_operator is not None:
            operator_str = self._operator_str(self._unary_operator)
        elif self._binary_operator is not None:
            operator_str = self._operator_str(self._binary_operator)
        else:
            raise ValueError(
                "Operator must be either unary or binary with a valid operator function"
            )
        res: dict[str, Any] = {
            "operator": operator_str,
        }
        res["left"] = (
            self._left._expression.__json__()
            if isinstance(self._left, ConfigVarExpression)
            else str(self._left)
        )
        if not self._is_unary and self._right is not None:
            res["right"] = (
                self._right._expression.__json__()
                if isinstance(self._right, ConfigVarExpression)
                else str(self._right)
            )
        return res

    def __str__(self) -> str:
        """String representation of this ConfigVarExpression.
        print the tree reprenting this expression
        """
        return json.dumps(self.__json__(), indent=2)



from __future__ import annotations

from typing import Any, Callable, Optional, Sequence, TypeGuard, Union, cast
from types import NotImplementedType
from pathlib import Path
from .utils import Undefined, _undef, PathType
from .config_var_def import ConfigVarDef
from .configvar_expression import (
    ConfigVarExpression,
    ConfigVarOperator,
    PathOperand,
    NumericOperand,
    StringOperand,
    UnaryOperator,
    BinaryOperator,
)
from numbers import Number
import operator as op


class ConfigVar(ConfigVarDef):
    """This class extends ConfigVarDef to add operations on ConfigVar for their composition"""

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)

    def _numeric_binary(
        self,
        other: PathOperand | NumericOperand | StringOperand,
        operator_fn: BinaryOperator,
    ) -> ConfigVarExpression:
        """Build a SegmentedNumeric binary expression with this ConfigVar as left operand."""
        return ConfigVarExpression(self, other, operator=operator_fn)

    def _numeric_rbinary(
        self,
        other: PathOperand | NumericOperand | StringOperand,
        operator_fn: BinaryOperator,
    ) -> ConfigVarExpression:
        """Build a SegmentedNumeric binary expression with this ConfigVar as right operand."""
        return ConfigVarExpression(other, self, operator=operator_fn)

    def __truediv__(
        self, other: PathOperand | NumericOperand | StringOperand
    ) -> ConfigVarExpression | NotImplementedType:
        """allow using operator / with ConfigVar on the left.
        The true nature of the operation that can be either a number division or a path composition is delayed until
        resolution time because the VarDef type_info is not known when the operator is applied.
        """
        return self._numeric_binary(other, op.truediv)

    def __rtruediv__(
        self,
        other: PathOperand | NumericOperand | StringOperand,
    ) -> ConfigVarExpression | NotImplementedType:
        """allow using operator / with ConfigVar on the right.
        The true nature of the operation that can be either a number division or a path composition is delayed until
        resolution time because the VarDef type_info is not known when the operator is applied.
        """

        return self._numeric_rbinary(other, op.truediv)

    def __add__(
        self, other: Number | ConfigVarDef | ConfigVarExpression
    ) -> ConfigVarExpression | NotImplementedType:
        """allow using operator + with ConfigVar on the left.
        The true nature of the operation that can be either a number addition or a string concatenation is delayed until
        resolution time because the VarDef type_info is not known when the operator is applied.
        """
        return self._numeric_binary(other, op.add)

    def __radd__(
        self, other: Number | ConfigVarDef | ConfigVarExpression
    ) -> ConfigVarExpression | NotImplementedType:
        """allow using operator + with ConfigVar on the right.
        The true nature of the operation that can be either a number addition or a string concatenation is delayed until
        resolution time because the VarDef type_info is not known when the operator is applied.
        """
        return self._numeric_rbinary(other, op.add)

    def __sub__(
        self, other: Number | ConfigVarDef | ConfigVarExpression
    ) -> ConfigVarExpression | NotImplementedType:
        """allow using operator - with ConfigVar on the left.
        The validity of the operation regarding operand type is delayed until
        resolution time because the VarDef type_info is not known when the operator is applied.
        """
        return self._numeric_binary(other, op.sub)

    def __rsub__(
        self, other: Number | ConfigVarDef | ConfigVarExpression
    ) -> ConfigVarExpression | NotImplementedType:
        """allow using operator - with ConfigVar on the right.
        The validity of the operation regarding operand type is delayed until
        resolution time because the VarDef type_info is not known when the operator is applied.
        """
        return self._numeric_rbinary(other, op.sub)

    def __mul__(
        self, other: Number | ConfigVarDef | ConfigVarExpression
    ) -> ConfigVarExpression | NotImplementedType:
        """allow using operator * with ConfigVar on the left.
        The validity of the operation regarding operand type is delayed until
        resolution time because the VarDef type_info is not known when the operator is applied.
        """
        return self._numeric_binary(other, op.mul)

    def __rmul__(
        self, other: Number | ConfigVarDef | ConfigVarExpression
    ) -> ConfigVarExpression | NotImplementedType:
        """allow using operator * with ConfigVar on the right.
        The validity of the operation regarding operand type is delayed until
        resolution time because the VarDef type_info is not known when the operator is applied.
        """
        return self._numeric_rbinary(other, op.mul)

    def __floordiv__(
        self, other: Number | ConfigVarDef | ConfigVarExpression
    ) -> ConfigVarExpression | NotImplementedType:
        """allow using operator // with ConfigVar on the left.
        The validity of the operation regarding operand type is delayed until
        resolution time because the VarDef type_info is not known when the operator is applied.
        """
        return self._numeric_binary(other, op.floordiv)

    def __rfloordiv__(
        self, other: Number | ConfigVarDef | ConfigVarExpression
    ) -> ConfigVarExpression | NotImplementedType:
        """allow using operator // with ConfigVar on the right.
        The validity of the operation regarding operand type is delayed until
        resolution time because the VarDef type_info is not known when the operator is applied.
        """
        return self._numeric_rbinary(other, op.floordiv)

    def __mod__(
        self, other: Number | ConfigVarDef | ConfigVarExpression
    ) -> ConfigVarExpression | NotImplementedType:
        """allow using operator % with ConfigVar on the left.
        The validity of the operation regarding operand type is delayed until
        resolution time because the VarDef type_info is not known when the operator is applied.
        """
        return self._numeric_binary(other, op.mod)

    def __rmod__(
        self, other: Number | ConfigVarDef | ConfigVarExpression
    ) -> ConfigVarExpression | NotImplementedType:
        """allow using operator % with ConfigVar on the right.
        The validity of the operation regarding operand type is delayed until
        resolution time because the VarDef type_info is not known when the operator is applied.
        """
        return self._numeric_rbinary(other, op.mod)

    def __pow__(
        self, other: Number | ConfigVarDef | ConfigVarExpression
    ) -> ConfigVarExpression | NotImplementedType:
        """allow using operator ** with ConfigVar on the left.
        The validity of the operation regarding operand type is delayed until
        resolution time because the VarDef type_info is not known when the operator is applied.
        """
        return self._numeric_binary(other, op.pow)

    def __rpow__(
        self, other: Number | ConfigVarDef | ConfigVarExpression
    ) -> ConfigVarExpression | NotImplementedType:
        """allow using operator ** with ConfigVar on the right.
        The validity of the operation regarding operand type is delayed until
        resolution time because the VarDef type_info is not known when the operator is applied.
        """
        return self._numeric_rbinary(other, op.pow)

    def __neg__(self) -> ConfigVarExpression | NotImplementedType:
        """Return unary negation expression for ConfigVar."""
        return ConfigVarExpression(
            self,
            None,
            operator=cast(UnaryOperator, op.neg),
            is_unary=True,
        )

    def __pos__(self) -> ConfigVarExpression | NotImplementedType:
        """Return unary plus expression for ConfigVar."""
        return ConfigVarExpression(
            self,
            None,
            operator=cast(UnaryOperator, op.pos),
            is_unary=True,
        )

    def __abs__(self) -> ConfigVarExpression | NotImplementedType:
        """Return absolute value expression for ConfigVar."""
        return ConfigVarExpression(
            self,
            None,
            operator=cast(UnaryOperator, op.abs),
            is_unary=True,
        )


def configvar(
    *,
    allow_override: Optional[bool] = None,
    auto_prolog: Optional[bool] = None,
    can_be_relative_to: Optional[Path | str] = None,
    default: Any | Callable[[str, Any, Any], Any] | Undefined = _undef,
    env_keys: Optional[str | Sequence[str]] = None,
    file_keys: Optional[str | Sequence[str]] = None,
    help: str = "",
    make_dirs: Optional[PathType] = None,
    mandatory: Optional[bool] = None,
    no_conffile_search: Optional[bool] = None,
    no_dir_processing: Optional[bool] = None,
    no_env_search: Optional[bool] = None,
    no_search: Optional[bool] = None,
    no_value_search: Optional[bool] = None,
    scopes: Optional[Sequence[str]] = None,
    split_to_list: Optional[bool | str] = None,
    transform: Optional[Union[Callable[[str, Any, Any], Any], str]] = None,
    type_info: Any = _undef,
    cli_keys: Optional[str | Sequence[str]] = None,
) -> Any:
    """Create a declarative configuration variable definition.

    Use it as a class attribute on an AppConfig subclass:

        class MyCfg(AppConfig):
            my_var: int = configvar(default=1)

    The variable fully qualified name defaults to the attribute name and its type is inferred from the type hint of the attribute declaration.

    This function allows to assign a ConfigVarDef objects to an attribute definitions of any type by telling the type checkers
    that the type of the assignation is Any instead of ConfigVarDef.
    Indeed ConfigVarDef is a generic class and my_var: int = ConfigVarDef(...) or my_var: ConfigVarDef[int] = ConfigVarDef(...)
    would not be accepted by type checkers. configvar() declares returning Any which deactivates type checking.

    """

    return ConfigVar(
        allow_override=allow_override,
        auto_prolog=auto_prolog,
        can_be_relative_to=can_be_relative_to,
        default=default,
        env_keys=env_keys,
        file_keys=file_keys,
        help=help,
        make_dirs=make_dirs,
        mandatory=mandatory,
        no_conffile_search=no_conffile_search,
        no_dir_processing=no_dir_processing,
        no_env_search=no_env_search,
        no_search=no_search,
        no_value_search=no_value_search,
        scopes=scopes,
        split_to_list=split_to_list,
        transform=transform,
        type_info=type_info,
        cli_keys=cli_keys,
    )



from __future__ import annotations

import jinja2
from typing import Any, Callable, Mapping, Optional, Sequence, Union, cast
from jinja2.nativetypes import NativeEnvironment
from .attr_dict import AttrDict
from .config_var_def import ConfigVarDef
from .utils import is_sequence, _undef
from pathlib import Path
from dataclasses import dataclass, field
from .configvar_expression import (
    ConfigVarExpression,
    config_path_expression,
    PathOperand,
    StringOperand,
    NumericOperand,
)


@dataclass
class SubstitutionResult:
    """
    A class to hold the result of a substitution operation.

    includes the rendered values of all candidate templates and the final rendered value.
    """

    rendered_value: Any = field(default=_undef)
    rendered_candidates: list[Any] = field(default_factory=list)


RenderContext = Mapping[str, Any]
SubstitutionType = Union[str, Path, ConfigVarDef, ConfigVarExpression]


class VarSubstituer:
    """
    A class to perform variable substitution on configuration variables.

    It resolve ConfigVarExpression and render Jinja2 templates.
    """

    # Native Jinja env: returns native Python objects when possible
    _JINJA_ENV: NativeEnvironment = NativeEnvironment(
        autoescape=False,
        undefined=jinja2.StrictUndefined,
    )

    # Filter style: {{ "/tmp" | P / "base" }}
    _JINJA_ENV.filters["p"] = config_path_expression

    # Global function style: {{ P("/tmp") / "base" }}
    _JINJA_ENV.globals["P"] = config_path_expression

    _var_solved: AttrDict

    def __init__(self, var_solved: AttrDict) -> None:
        """Create a VarSubstituer with the provided context for variable substitution.
        var_solved must be dynamic and contain the currently resolved values."""
        self._var_solved = var_solved

    def render_jinja_native(self, template_value: str) -> Any:
        """Render a template with NativeEnvironment and return native objects."""
        template = self._JINJA_ENV.from_string(template_value)
        return template.render(self._var_solved.to_dict())

    def do_substitution(
        self,
        value: Any,
    ) -> Any:
        """
        Render a value by substituing configuration variables with their value and rendereding jinja2 template.



        The value can be:
        - If the value is not a string, nor a Path, nor a ConfigVarDef, nor a ConfigVarExpression nor a dict or a sequence , it is returned as is.
        - If the value is a string it is rendered with Jinja, as we use Native Jinja environment the result can be a Python type that can need substitution itself,
            so we pass the redered value down to test all other cases. (Jinja can't return a jinja string.)
        - If the value is a ConfigVarExpression it is resolved and with the result we do a recurvise call in order to test all cases including the case of a
              Jinja template resulting from the expression resolution.
        - If the value is a Path it is returned as a Path.
        - If the value is a sequence (except strings), each item is resolved recursively and the result is returned as the same sequence type.
        - If the value is a dict, each value is rendered recursively and the result is returned as a dict with the same keys and the rendered values.
        - If the value is a  ConfigVarDef then look in the context if it is already defined, if it is the case the value is resolve recursively and returned,
          otherwise raise an exception.
        """

        val_to_render: Any = value
        if value is None:
            return None
        if isinstance(value, str):
            try:
                # we can't do a recursive call because we will loop on the string case.
                # instead we continue the process with the rendered value
                val_to_render = self.render_jinja_native(value)
            except jinja2.exceptions.UndefinedError as e:
                raise ValueError(f"Undefined variable in template: {value}") from e
            except jinja2.TemplateError as e:
                raise ValueError(
                    f"Error rendering template: {value}. Error: {str(e)}"
                ) from e
        if isinstance(val_to_render, Path):
            res = self.do_substitution(str(val_to_render))
            return Path(res) if res is not None else None

        elif isinstance(val_to_render, ConfigVarDef):
            if val_to_render.fq_name is None or not self._var_solved.has_value(
                val_to_render.fq_name
            ):
                raise ValueError(
                    f"Variable {val_to_render.fq_name} is not defined. Cannot render ConfigVarDef without a value."
                )
            return self.do_substitution(self._var_solved[val_to_render.fq_name])
        elif isinstance(val_to_render, ConfigVarExpression):
            res = val_to_render.resolve(self.do_substitution)
            return res
        elif is_sequence(val_to_render):
            return type(val_to_render)([self.do_substitution(item) for item in val_to_render])  # type: ignore
        elif isinstance(val_to_render, dict):
            return type(val_to_render)(
                {k: self.do_substitution(v) for k, v in val_to_render.items()}
            )
        else:
            return val_to_render

    def resolve_candidate_list(
        self,
        *,
        candidates: SubstitutionType | None | Sequence[SubstitutionType],
        check_exists: Callable[[str], Any],
    ) -> SubstitutionResult:
        """
        Takes a list of candidates values (file paths,env name, value key or file_key) and resolves them in
        order, returning the first one for which check_exists returns a non-_undef value.

        Args:
            candidates: A single string, a Path, a ConfigVarDef or a ConfigVarExpression, or a list of these. Values can be
            potential jinja templates for string.
            check_exists: An predicate to validate the rendered string.
                        returns _undef if not valid, or the value to return if valid.
        Returns:
            A SubstitutionResult containing the rendered candidates and the first rendered value that passed the
            check_exists validation. If no candidate is valid, rendered_value will be _undef.
        """
        res = SubstitutionResult()
        if candidates is None:
            return res

        candidate_list: list[SubstitutionType] = []

        if is_sequence(candidates):
            candidate_list = cast(list[SubstitutionType], candidates)
        else:
            candidate_list = [cast(SubstitutionType, candidates)]

        for raw_candidate in candidate_list:
            rendered = self.do_substitution(raw_candidate)
            res.rendered_candidates.append(rendered)
            valid = check_exists(rendered)
            if valid != _undef:
                res.rendered_value = valid
                return res

        return res



from __future__ import annotations

import logging
import os
from typing import Any, Callable, Iterable, Mapping, Optional, Sequence, cast
from .config_var_def import ConfigVarDef, Flags
from .attr_dict import AttrDict
from .utils import AppConfigException, Undefined, _undef, is_sequence, PathType
from .var_substitution import VarSubstituer
from .configvar_expression import PathOperand
from pydantic import TypeAdapter, ValidationError
from pathlib import Path


class Solver:
    """Class responsible for solving configuration variables values
    according to the defined search order and applying transformations and validations."""

    _logger = logging.getLogger(__name__)
    _local_flags: Flags
    _values_map: Mapping[str, Any]
    """ The values provided directly to the AppConfig instance."""
    _conffile_values: Optional[AttrDict] = None
    """ The values found in the configuration file."""
    _eval_runner: Callable[[Callable[[str, Any, Any], Any], str, Any], Any]
    """ The function that will run the evaluation of callables for default values and transformations."""
    _var_solved: AttrDict
    """ currently solved config var. Used to handle dependencies between variables and variable substitution """
    _var_substituer: VarSubstituer
    """ The VarSubstituer instance used to substitute variables in the value found for a variable to solve."""

    def __init__(
        self,
        *,
        flags: Flags,
        values_map: Mapping[str, Any],
        eval_runner: Callable[[Callable[[str, Any, Any], Any], str, Any], Any],
        var_solved: AttrDict,
        conffile_values_map: Optional[Mapping[str, Any]] = None,
    ) -> None:
        self._local_flags = flags
        self._values_map = values_map
        self._conffile_values = self.set_conffile_values_map(conffile_values_map)
        self._eval_runner = eval_runner
        self._var_solved = var_solved
        self._var_substituer = VarSubstituer(self._var_solved)

    def set_conffile_values_map(
        self, conffile_values_map: Optional[Mapping[str, Any]]
    ) -> None:
        """
        Set or update the configuration file values map.
        Allow to use the same solver with ou without configuration file.
        """
        if conffile_values_map is not None:
            self._conffile_values = AttrDict(conffile_values_map)

    @classmethod
    def _find_var_in_dict(
        cls, where: Mapping[Any, Any], var_name: str
    ) -> Any | Undefined:
        """
        Search for var_name in a dictionary.
        var_name contains dict keys separated by a dot (.) ie key1.key2.key3
        The left side key key1 is search in the where directory, if it is found
        and it is a dictionary then key2 is search in it and
        all key are searched in sequence.
        Returns _undef if not found, allowing to distinguish between None (explicit null) and missing.
        """
        cur_dic: Mapping[Any, Any] = where
        for key in var_name.split("."):
            # logger = logging.getLogger(f"{cls.__module__}.{cls.__name__}")
            # logger.debug(f"searching for key: {key} in dict : {cur_val}")
            if key in cur_dic:
                cur_dic = cur_dic[key]
                # logger.debug(f"Found key: {key} = {cur_val}")
            else:
                return _undef

        return cur_dic

    def resolve_conffile_path(
        self,
        candidates: Optional[PathOperand | Iterable[PathOperand]],
    ) -> Path | Undefined:
        """
        Resolve the configuration file path from candidates, returns the Path corresponding to the first valid canditate.

        Each candidate can be a ConfigVarDef, a ConfigVarExpression or a string that can be a be resolved with the already resolved variables in this AppConfig.
            - A string is rendered with jinja and transform to a Path
            - A Path is stringified and rededered with jinja, then transform back to a Path
            - A ConfigVarExpression is resolved and returned with its corresponding type and then processed according to its type as above.
            - A ConfigVarDef is resolved to its value and then processed according to its type as above.

        Candidates are valid if a file exists at their rendered path, if no valid candidate is found then the rendered_value attribute of
        the returned object is set to _undef otherwise it contains the the resolution of the first valid candidate.

        The check_exists function check that the file exists at the resolved path and return the path if it exists or _undef if it doesn't exist.

        The validation checks that the file exists at the resolved path. This allows to have multiple
        potential location for the configuration file and use the first one found.
        """
        if candidates is None:
            return _undef

        # Normalize candidates to list of str for resolve_candidate_list
        candidates_list: list[PathOperand] = []
        if not is_sequence(candidates):
            candidates_list = [cast(PathOperand, candidates)]
        else:
            candidates_list = cast(list[PathOperand], candidates)

        # Define existence check for file paths
        def file_exists(p: str | Path) -> Path | Undefined:
            path_obj: Path
            if isinstance(p, str):
                path_obj = Path(p)
            elif isinstance(p, Path):
                path_obj = p
            else:
                raise ValueError(
                    f"Invalid candidate type: {type(p)}. Expected str or Path."
                )

            if path_obj.is_file():
                return path_obj
            else:
                return _undef

        resolution_status = self._var_substituer.resolve_candidate_list(
            candidates=candidates_list, check_exists=file_exists
        )

        return resolution_status.rendered_value

    def resolve_confvar(
        self, *, var_to_solve: ConfigVarDef, scopes: Optional[str | Iterable[str]]
    ) -> Any:
        """
        Assign a value to the given ConfigVarDef according to the search order:
        1. values mapping
        2. environment variables
        3. configuration file
        4. default value
        Applies transformations and validations as specified in the ConfigVarDef.
        If its a Path variable, process relative paths, create directories if needed.

        Returns _undef if no value have been found and the variable is not mandatory or if the variable is out of the given scopes.
        Raises AppConfigException if the variable is mandatory and no value have been found or if validation failed.
        """
        val_found: Any = _undef
        found = False

        cur_flags = self._local_flags.merge(inplace=False, other=var_to_solve.flags)
        assert cur_flags is not None

        # if we are resolving the variables in some specific scopes (ie parameter scope has value)
        # we check if the variable belongs to one of these scopes and if not we skip it, in this
        # case the variable is said to be "out of scope" and has no existence.
        if scopes is None and var_to_solve.scopes is not None:
            return _undef
        # if we are resolving the variable in some specific scopes (ie parameter scope has value) and the variable
        # has also some scopes defined we check if there is at least one scope in common
        if scopes is not None and var_to_solve.scopes is not None:
            test_ctx = []
            # put True in test_ctx for each scope that matches, False otherwise
            if is_sequence(scopes):
                test_ctx = [c in var_to_solve.scopes for c in scopes]
            else:
                test_ctx = [scopes in var_to_solve.scopes]

            # test if there is any True in test_ctx which means that at least one scope matches
            if not any(test_ctx):
                return _undef
        # if the variable has no scope and we are resolving in some specific scopes this means that
        # the variable is common to all scopes so we process it but only once except if the flag
        # allow_override is set to True
        if var_to_solve.scopes is None and scopes is not None:
            if not cur_flags.allow_override and var_to_solve.fq_name in self._var_solved:
                return _undef

        # reading fq_name and reassure type checker that it has a value.
        var_fq_name = var_to_solve.fq_name
        if var_fq_name is None:
            raise AppConfigException("Internal error: var_to_solve.fq_name is None")

        #  Searching variable value in the values mapping
        if (
            self._values_map is not None
            and not cur_flags.no_value_search
            and var_to_solve.cli_keys is not None
        ):
            # in the value map we have directly the variable fq_name as key thanks to the click option callback
            # that restore the fq_name instead of the value_keys.
            if (
                var_fq_name in self._values_map
                and self._values_map[var_fq_name] != _undef
            ):
                val_found = self._values_map[var_fq_name]
                found = True
                self._logger.debug(
                    f"{var_fq_name} -> Found in Values with fq_name as key = {val_found}"
                )

        # Searching variable value in the environment variables
        if (
            not found
            and not cur_flags.no_env_search
            and var_to_solve.env_keys is not None
        ):

            def check_env(k: str) -> Any:
                """check if the key k exists in the environment variables and return its value or _undef"""
                res = os.getenv(k)
                return _undef if res is None else res

            # calling the substituer to resolve the env_keys with variable substitution and jinja rendering, we pass check_env as
            # check_exists to stop at the first existing env key and return its value.
            resolution_status = self._var_substituer.resolve_candidate_list(
                candidates=var_to_solve.env_keys, check_exists=check_env
            )
            value_checked = resolution_status.rendered_value

            if value_checked != _undef:
                val_found = value_checked
                found = True
                self._logger.debug(f"{var_fq_name} -> Found in Env = {val_found}")

        #  Searching variable value in the configuration file
        if (
            not found
            and not cur_flags.no_conffile_search
            and self._conffile_values is not None
            and var_to_solve.file_keys is not None
        ):

            def check_file(k: str) -> Any:
                """check if the key k exists in configuration file and return its value or _undef"""
                if self._conffile_values is None:
                    return False
                return (
                    _undef
                    if not self._conffile_values.has_value(k)
                    else self._conffile_values[k]
                )

            # calling the substituer to resolve the file_keys with variable substitution and jinja rendering, we pass check_file as
            # check_exists to stop at the first existing file key and return its value.
            resolution_status = self._var_substituer.resolve_candidate_list(
                candidates=var_to_solve.file_keys, check_exists=check_file
            )
            value_checked = resolution_status.rendered_value
            # None is a valid value from config file, _undef means not found
            if value_checked != _undef:
                val_found = value_checked
                found = True
                self._logger.debug(
                    f"{var_fq_name} -> Found in Configuration File = {val_found}"
                )

        # Setting variable value to the default value if defined. The eval form can return _undef
        # to indicate that there is no default value to use, allowing to distinguish between no default value
        # and an explicit default value of None.
        # Even when allow_override is True  default value can't override an already found value (a value found in a previous call to resolve_vars()).
        if (
            var_fq_name not in self._var_solved
            and not found
            and var_to_solve.default is not _undef
        ):
            if var_to_solve.default is not None and callable(var_to_solve.default):
                val_found = self._eval_runner(var_to_solve.default, var_fq_name, None)
            else:
                val_found = var_to_solve.default
            found = val_found != _undef
            self._logger.debug(f"{var_fq_name} -> Found in Default Value = {val_found}")

        # Raise exception if no value was found and variable is mandatory and if we have not
        # already a value for this variable (in case of override with scopes for example)
        if not found and var_to_solve.mandatory and var_fq_name not in self._var_solved:
            raise AppConfigException(f"No value for var {var_fq_name} in scope {scopes}")

        if found:
            if isinstance(val_found, Undefined):
                raise AssertionError(
                    f"Internal error on var {var_fq_name}: val_found is _undef while found=True"
                )
            # spliting lists
            if (
                var_to_solve.split_to_list
                and val_found is not None
                and isinstance(val_found, str)
            ):
                sep: str = (
                    ","
                    if isinstance(var_to_solve.split_to_list, bool)
                    else var_to_solve.split_to_list
                )
                val_found = val_found.split(sep)

            # Transform the variable value if specified, we do it before rendering because the transformation can produce a value
            # that need to be rendered and also because the transformation can produce a non string value that we don't want to render.
            if var_to_solve.transform is not None:
                val_transfo = self._eval_runner(
                    cast(Callable, var_to_solve.transform),
                    var_fq_name,
                    val_found,
                )
                self._logger.debug(
                    f"{var_fq_name} -> Value Transformed: {val_found} => {val_transfo}"
                )
                val_found = val_transfo

            # Apply variable substitution and Jinja template rendering to the found value using the currently solved variables.
            # we delegate to the var_substituer the handling of different type of value and the rendering or not according to the type.
            if val_found is not None:

                if is_sequence(val_found):
                    # if we have a list of value we render each value separately
                    val_found = [
                        self._var_substituer.do_substitution(v)
                        for v in cast(Iterable, val_found)
                    ]
                else:
                    val_found = self._var_substituer.do_substitution(val_found)

            # validate and cast the variable value using pydantic TypeAdapter
            if var_to_solve.type_info is not None and val_found is not None:
                try:
                    ta = TypeAdapter(
                        var_to_solve.type_info,
                        config={"coerce_numbers_to_str": True, "strict": False},
                    )
                    val_found = ta.validate_python(val_found)
                except ValidationError as e:
                    raise AppConfigException(
                        f"Validation failed for var {var_fq_name}: {e}"
                    )

            # Make special treatment for paths
            if (
                not var_to_solve.no_dir_processing
                and var_to_solve.type_info == Path
                and val_found is not None
            ):
                new_val: list[Path] | Path | None = None

                val_found = self._process_paths(
                    value=val_found, var=var_to_solve, var_solved=self._var_solved
                )

        return val_found

    def _process_paths(
        self, *, value: Any, var: ConfigVarDef, var_solved: AttrDict
    ) -> Path:
        """
        Run the path processing for a single Path variable.
        """
        var_value: Path
        if isinstance(value, str):
            var_value = Path(value)
        elif isinstance(value, Path):
            var_value = value
        else:
            raise Exception("not a path")

        res: Path = var_value
        # First process the can_be_relative_to case.
        if var.can_be_relative_to is not None:

            if var_value.is_absolute():
                res = var_value
            else:
                # resolving the root directory from which to be relative to.
                can_be_relative = var.can_be_relative_to

                root_dir: Path
                if isinstance(can_be_relative, ConfigVarDef):
                    can_be_relative = can_be_relative.fq_name
                if can_be_relative is None:
                    raise AppConfigException(
                        f"Internal error: can_be_relative_to is None for var {var.fq_name}"
                    )
                if (
                    isinstance(can_be_relative, str)
                    and can_be_relative in var_solved
                    and var_solved[can_be_relative] != _undef
                ):
                    root_dir = Path(var_solved[can_be_relative])
                else:
                    root_dir = Path(can_be_relative)
                res = root_dir.joinpath(var_value)

        # Expanding user home and resolving dots in path
        res = res.expanduser().resolve()

        # create directory if make_dirs is not None
        if var.make_dirs:
            if var.make_dirs == PathType.Dir:
                res.mkdir(parents=True, exist_ok=True)
            else:
                res.parent.mkdir(parents=True, exist_ok=True)
        return res



from .app_config import AppConfig
from .config_var_def import ConfigVarDef
from .config_var import configvar, ConfigVar
from .click_option import click_option
from .utils import PathType, AppConfigException, NameSpace
from .configvar_expression import ConfigVarExpression, config_path_expression
from .attr_dict import AttrDict

Sp = config_path_expression
Ce = ConfigVarExpression

__all__ = [
    "AppConfig",
    "AppConfigException",
    "ConfigVarDef",
    "NameSpace",
    "configvar",
    "ConfigVar",
    "click_option",
    "PathType",
    "Sp",
    "Ce",
    "AttrDict",
]



from __future__ import annotations

from typing import Any, Callable, cast

import click
from click.types import BoolParamType

from the1conf import ConfigVarDef
from the1conf.utils import Undefined, _undef


class StringOrUndefined(click.ParamType):
    name = "string"

    def convert(self, value: Any, param: Any, ctx: Any) -> Any:
        if value is _undef:
            return value
        return click.STRING.convert(value, param, ctx)


class BoolOrUndefined(BoolParamType):
    name = "boolean"

    def convert(self, value: Any, param: Any, ctx: Any) -> Any:
        if value is _undef:
            return value
        return super().convert(value, param, ctx)


def click_option(config_var: Any, **kwargs: Any) -> Callable[[Any], Any]:
    """Wrapper for click.option with the metadata from ConfigVarDef."""
    if not isinstance(config_var, ConfigVarDef):
        raise TypeError(f"click_option expects a ConfigVarDef, got {type(config_var)}")

    # setting the option names (--foo, -f) and flags for boolean variables based on the config variable.
    param_decls = []

    keys = config_var.cli_keys
    if keys is None or len(keys) == 0:
        raise ValueError(
            f"Config variable {config_var.fq_name} has no value_key, which means it won't be settable via click options. This is likely a misconfiguration of the variable, check that it is not defined with no_search flags."
        )
    is_bool = config_var.type_info is bool

    for key in keys:
        if len(key) == 1:
            param_decls.append(f"-{key}")
        else:
            if is_bool:
                param_decls.append(f"--{key}/--no-{key}")
            else:
                param_decls.append(f"--{key}")

    # Used as destination name in Click internals.
    # We build it from fq_name and replace namespace separators (.) with
    # the unicode character ː (U+02D0) so Click can infer a valid option name.
    if config_var.fq_name is None:
        raise ValueError(
            "Config variable fq_name is not set. Make sure to define click options on class attributes with a resolved ConfigVarDef."
        )
    param_name = config_var.fq_name.replace(".", "\u02d0")

    if is_bool:
        kwargs["is_flag"] = True

    # Add help.
    if "help" not in kwargs and config_var.help:
        kwargs["help"] = config_var.help

    # We overwrite any passed type to ensure that resolve_vars receives a string.
    # We use a custom type to handle the `_undef` default value without converting
    # it to string.
    if not is_bool:
        kwargs["type"] = StringOrUndefined()
    else:
        kwargs["type"] = BoolOrUndefined()

    # We set default to `_undef` to distinguish between "option not provided" (_undef)
    # and "option provided with no value" (which shouldn't happen with string) or None.
    kwargs["default"] = _undef

    # Add default in help message.
    if (
        "show_default" not in kwargs
        and config_var.default is not Undefined
        and not callable(config_var.default)
    ):
        # We must convert to string otherwise click interprets bool/int as flags (True/False)
        # which tell it "show the default" (which is None here), so it displays nothing.
        kwargs["show_default"] = str(config_var.default)

    def _get_the1conf_click_callback(
        var_fq_name: str, user_callback: Callable[[Any, Any, Any], Any]
    ) -> Callable[[Any, Any, Any], Any]:
        """Parameters (options and arguments) are forwarded to the command callbacks.
        One common way to prevent a parameter from being passed to the callback is set the click argument `expose_value` to False.
        This will hide the parameter entirely so that it is not passed to the callback and also not stored in the Click context params.

        This allows us to store the value in the Click context with the original the1conf key (with dots and '-') instead of the click option name
        (with dots and '-' replaced), which makes it easier to resolve the variable in resolve_vars.
        """
        var_fq_name = var_fq_name
        user_callback = user_callback

        def call_back(ctx: Any, param: Any, value: Any) -> Any:
            ctx.params[var_fq_name] = value
            if callable(user_callback):
                return user_callback(ctx, param, value)
            return value

        return call_back

    # we are going to use a callback and we don't want to override any user provided callback,
    # so we store the user callback to call it later in our own callback
    user_callback = cast(Callable[[Any, Any, Any], Any], kwargs.get("callback"))

    kwargs["callback"] = _get_the1conf_click_callback(config_var.fq_name, user_callback)
    # We set expose_value to False to prevent Click from storing the value in the context with the
    # option name (with dots replaced), which would be different from the original
    # the1conf key and would make it harder to resolve the variable in resolve_vars.
    kwargs["expose_value"] = False

    def _decorator(func: Any) -> Any:
        """this is a docorator function that takes as parameter the function being decorated.

        It is used on a click command function to intercept the click option added and verify that there
        are no duplicate option names.
        """
        existing_option_names: set[str] = set()
        # read the __click_params__ of click that contains the list of options and arguments already declared.
        # we keep on the click.Option parameter (not click.Argument or other things).
        for param in getattr(func, "__click_params__", []):
            if isinstance(param, click.Option):
                existing_option_names.update(param.opts)

        # expand boolean parameter  "--feature/--no-feature" in param_decls (get from the closure)
        # by splitting them in two.
        expanded_param_decls = {
            token for declaration in param_decls for token in declaration.split("/")
        }
        # test duplication. We check that none of the new option names (after expansion)
        # are already declared in the function.
        duplicate_option_names = sorted(
            existing_option_names.intersection(expanded_param_decls)
        )
        if duplicate_option_names:
            raise ValueError(
                "Duplicate click option declaration detected: "
                f"{', '.join(duplicate_option_names)}. "
                "Each CLI option key must be unique across config variables."
            )

        return click.option(*param_decls, param_name, **kwargs)(func)

    # Note: We do NOT pass 'envvar' nor 'default'.
    # The callback stores values with the original the1conf key in ctx.params.
    return _decorator



from __future__ import annotations

from typing import Any, Iterable, Mapping, Optional, cast, Iterator
from .config_var import ConfigVar
from .utils import get_attribute_docstrings, NameSpace


class ConfigVarCollector:
    """
    Collector class to gather ConfigVar from classes and their nested NameSpace classes.
        This class is used in the AppConfigMeta metaclass to collect ConfigVar defined in the various base
        classes and merge them to create the _config_var_defs
        It acts as a context manager to let the AppConfigMeta metaclass decide which classes are to be scanned
        and in which order. The meta class will call the collect_config_vars on as many classes as needed and will
        get the collected variables at the end of its scanning process.

        collect_config_vars is the main method:
        - it first loop over the attributes declared in the targeted class by scanning its __dict__.
            - when it finds a ConfigVar:
              - it adds it to the collected variables
              - attach the namespace and fq_name to the ConfigVar instance.
              - If the ConfigVar instance doesn't have help defined it tries to use the attribute
              docstring as help.

        - Then it merges the collected variables with the already collected variables from its previous runs.
        - The merge strategy is the following:
          - We need to merge the collected ConfigVar attributes found in the various classes and keep the order of their definition,
            to do so:
            - we loop over the classes _config_var_defs in reversed MRO class order from the furthest ancestor to the closest parent to
              keep python order of class definition (last definition wins).
                - when we discover an new attribute not yet defined we add it at the last position.
                - when discovering an attribute that was already defined in an ancestor we remove the old one and add the new one its new
                position at the end of the list.
            - For Namespaced attribute (defined in a nested NameSpace class) we add them in the same way as other attributes, collisions
              are detected by accounting for the namespace path. Conflicts are resolved in the same way as other attributes. The main difference
              is when a namespace is masked by a non-namespace attribute defined later in the MRO, in this case we need to remove all
              the attributes defined in the masked namespace (including nested namespaces).
    """

    _collected_vars: dict[str, ConfigVar]
    """  The collected ConfigVar indexed by their fq_name. This is the main output of the collector, 
         it will be used to create the _config_var_defs attribute of the AppConfig class."""
    _collected_namespaces: dict[tuple[str, ...], list[str]]
    """ A dict with all the namespaces that have been collected along with the list of the variable fq_name they contain, this is used to detect 
        namespace masking when merging new variables."""
    _auto_prolog_var_names: list[str]
    """ The list of the names of ConfigVar whose attribute auto_prolog is True, this is used to know which variables to put in the prolog file when calling save_prolog."""

    def __enter__(self):
        """Enter the runtime context related to this object."""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Exit the runtime context and clean up resources."""
        self.close()

    def __init__(self) -> None:
        """Create a new ConfigVarCollector instance."""
        self._collected_vars = {}
        self._collected_namespaces = {}
        self._auto_prolog_var_names = []

    def close(self) -> None:
        """Clean up resources if needed."""
        self._collected_vars.clear()
        self._collected_namespaces.clear()
        self._auto_prolog_var_names.clear()

    def collect_config_vars(self, owner: type, ns: tuple[str, ...] = ()) -> None:
        """Entry point to collect ConfigVar from a class and its nested NameSpace classes.
        Get the attributes and merge them.
        """

        conf_var_list: Iterable[ConfigVar]
        if "_config_var_defs" in owner.__dict__:
            # If the owner class already has _config_var_defs, it means it's a subclass of AppConfig and we can just
            # merge its _config_var_defs
            inner_list = cast(
                dict[str, ConfigVar], getattr(owner, "_config_var_defs", None)
            )
            assert inner_list is not None
            conf_var_list = inner_list.values()
        else:
            # If the owner class doesn't have _config_var_defs, it means it's not a subclass of AppConfig and we need to scan it for ConfigVarDef attributes
            conf_var_list = self._extract_config_vars(owner, ns=ns)

        self._merge_config_var_defs(conf_var_list)

    def _extract_config_vars(
        self, owner: type, ns: tuple[str, ...] = ()
    ) -> Iterable[ConfigVar]:
        """
        Search for ConfigVar defined in the owner class and its nested NameSpace classes.

        loop over the class __dict__ to find ConfigVar defined in this class and yield them.

        If a ConfigVar doesn't have help defined we try to attach the docstring of the attribute as help if it exists.

        Add its namespace to the ConfigVar instance.
        Add its _fq_name to the ConfigVar instance.
        """
        # extract docstrings from the class source code
        docstrings = get_attribute_docstrings(owner)

        # Inspect the class __dict__ to find ConfigVar defined in this class
        for attr_name, value in owner.__dict__.items():
            if isinstance(value, ConfigVar):
                # If help is missing, try to use the docstring
                if not value.help and attr_name in docstrings:
                    value.help = docstrings[attr_name]

                name = getattr(value, "name", None)
                if name is None:
                    raise ValueError(
                        f"attribute {attr_name} of class {owner.__name__} is missing attribute 'name' in its ConfigVar, this should never happen."
                    )
                value.fq_name = f"{'.'.join((*ns, name))}"
                value.name_space = ns
                yield value

            elif (
                isinstance(value, type)
                and issubclass(value, NameSpace)
                and value is not NameSpace
            ):
                # Recurse into nested NameSpace classes
                yield from self._extract_config_vars(value, ns=(*ns, attr_name))

    def _merge_config_var_defs(self, new_config_var_defs: Iterable[ConfigVar]) -> None:
        """
        Merge new ConfigVar attributes into the already collected one.

        The merge strategy is explain at the class level docstring
        """
        if not new_config_var_defs:
            return

        for var_def in new_config_var_defs:
            # getting data from var_def and check they are OK.
            var_namespace = cast(tuple[str, ...], getattr(var_def, "name_space", None))
            var_fq_name = cast(Optional[str], getattr(var_def, "fq_name", None))
            if var_namespace is None or var_fq_name is None:
                raise ValueError(
                    f"ConfigVar {var_def.fq_name} is missing _name_space attribute, this should never happen."
                )
            var_fq_name_tuple = tuple(var_fq_name.split("."))

            # first we check if we are masking a namespace.
            if var_fq_name_tuple in self._collected_namespaces.keys():
                # we are masking a namespace, we need to remove all the variables defined in this namespace and its sub-namespaces
                ns_to_remove = []
                for ns in self._collected_namespaces.keys():
                    if ns[: len(var_fq_name_tuple)] == var_fq_name_tuple:
                        ns_to_remove.append(ns)

                # remove the variables defined in the masked namespaces
                for ns in ns_to_remove:
                    for var_to_remove in self._collected_namespaces.get(ns, []):
                        self._collected_vars.pop(var_to_remove)
                        if var_to_remove in self._auto_prolog_var_names:
                            self._auto_prolog_var_names.remove(var_to_remove)
                    # removing the name spaces
                    self._collected_namespaces.pop(ns, None)

            # check if the new var_def is already in the collected vars
            if var_fq_name in self._collected_vars:
                # Override previous definition
                self._collected_vars.pop(var_fq_name)
                if var_fq_name and var_def.auto_prolog:
                    self._auto_prolog_var_names.remove(var_fq_name)

            if var_namespace not in self._collected_namespaces:
                # test if we have found a new namespace, note that empty tuple is a valid namespace for root variable.
                self._collected_namespaces[var_namespace] = []

            # the new var_def is added at the end of the list to preserve the order
            self._collected_vars[var_fq_name] = var_def
            self._collected_namespaces[var_namespace].append(var_fq_name)

            auto_prolog = getattr(var_def, "auto_prolog", None)
            if auto_prolog:
                self._auto_prolog_var_names.append(var_fq_name)

    def get_collected_vars(self) -> dict[str, ConfigVar]:
        """
        Get the collected configuration variables.
        """
        return self._collected_vars.copy()

    def get_auto_prolog_var_names(self) -> list[str]:
        """
        Get only the collected configuration variables that have auto_prolog enabled.
        """
        return self._auto_prolog_var_names.copy()


class AppConfigMeta(type):
    """
    Metaclass for AppConfig and its subclasses to handle class initialization.

    Specifically it only defines the __init__() method with the objective to collect configuration variables
    from AppConfig and its subclasses.

    It set the _config_var_defs attribute of the initialized class with as a dictionary that contains all the ConfigVar
    attributes defined in the class and its ancestrors (parents and parents of parents) indexed by their
    fully qualified name (fq_name), which is the name prefixed with its namespaces separated by dots.

    The order of the attributes is the same as the order of their definition in the hierarchy of classes from the deepest
    ancestor first to the class itself last.The order in which the classes are processed follow the MRO order of the classes being initialized.
    In case of conflict the last definition in the MRO order wins, the older one is removed.
    *Note*: if a NameSpace is masked by a non-namespace attribute defined later in the MRO, all the variables defined in this
    namespace and its sub-namespaces are removed.

    It also set the _autoprolog_var_names attribute of the initialized class with the list of the names of the ConfigVar attributes that have auto_prolog enabled,
    this is used to know which variables to put in the prolog processing when calling resolve_vars().

    During the collection the metaclass also attach the to the ConfigVar instance their namespace and their fq_name.
    It also tries to use the attribute docstring as help if the help is missing.

    """

    def __init__(cls, name: str, bases: tuple[type, ...], attrs: dict[str, Any]) -> None:
        """
        Python metaclass method called when a new class belonging to this metaclass is created. It is called on direct subclasses
        and subclasses of those subclasses.

        - Difference between __mro__ and __bases__
            cls.__bases__ : the direct parents (what you wrote in the definition).
            cls.__mro__ : the complete method resolution order (parents + ancestors) in the effective order.

        Synopsis of the collection strategy:

        - We are collecting the ConfigVar attributes defined in the base classes and merge them to create the _config_var_defs
          list of the current class.

        The collection strategy is the following:
        - If the base has no _config_var_defs attributes it means that it's not a subclass of AppConfig and we need to scan its MRO
          for  searching for ConfigVar attributes defined in its ancestors. In this case we scan all ancestors by starting from
          the deepest ancestor to the closest parent in reversed MRO and we apply the merge strategy described above to merge the
          collected ConfigVar attributes.
        - If a base class has the _config_var_defs attribute it means that it is a subclass of AppConfig and that we have
          already collected its ConfigVar attributes. In this case we can just merge the base class _config_var_defs attribute
          instead of scanning the class again.

        Args:
            name (str): The name of the class being created.
            bases (tuple[type, ...]): A tuple containing the base classes of the class being created.
            attrs (dict[str, Any]): A dictionary containing the class namespace (attributes and methods)
                                  populated during the execution of the class body.
                                  Keys are the names of the attributes (strings) and values are the
                                  attribute values (methods, class variables, properties, etc.).
        """

        super().__init__(name, bases, attrs)

        # we use the ConfigVarCollector to collect the ConfigVar attributes defined in the various classes.
        # It also handle the merge strategy and the namespace masking detection.
        with ConfigVarCollector() as collector:
            for base in reversed(bases):
                if not "_config_var_defs" in cls.__dict__:
                    # If the base class has not a _config_var_defs, it means it's not a subclass of AppConfig and we need
                    # to collect the ConfigVar by scanning its ancestors, note that the first class in _mro_ is the class itself
                    # so we skip and we also skip the last one which is object.
                    for ancestor in reversed(base.__mro__[1:-1]):
                        # we need to detect if we have passed an AppConfig class in the MRO.
                        # to avoid scanning parent that already have their ConfigVar collected and stored in their _config_var_defs attribute.
                        # Problem 1: we can't import AppConfig because we are building it.
                        # Problem 2: the MRO contains also `lateral` classes in case of multiple inheritance so it's not possibl to say that when we
                        #   a processing a class whose name is 'AppConfig' all remaining classes will be subclasses of AppConfig (if they are lateral inheritance)

                        collector.collect_config_vars(ancestor)

                # in all case we need to collect the ConfigVar defined in the base class itself:
                #   - if it has a _config_var_defs attribute the collector we will merge it.
                #   - if it doesn't have a _config_var_defs the collector will collect the ConfigVar defined in its __dict__.
                collector.collect_config_vars(base)

            # finally we collect the ConfigVar defined in the initialized class itself
            collector.collect_config_vars(cls)

            cls._config_var_defs = collector.get_collected_vars()
            cls._autoprolog_var_names = collector.get_auto_prolog_var_names()
